"""The `itch` module contains the code necessary to push projects to Itch.io using the butler
    tool.
"""
import subprocess as proc
import enum


class ButlerPlatformType(enum.Enum):
    """Enumerations for platform types."""
    WINDOWS = "win"
    DARWIN = "mac"
    LINUX = "linux"
    OTHER = ""


DEFAULT_TAG_RULES = {
    "win": ButlerPlatformType.WINDOWS,
    "mac": ButlerPlatformType.DARWIN,
    "linux": ButlerPlatformType.LINUX
}


class Butler(object):
    """The handler for publishing content to Itch.io using Butler.

    Attributes:
        author (str): The itch.io username that is publishing content.
        project (str): The project that the author is publishing content for.
        bin (str): The path to the Butler executable.
        tag_rules (dict): A dictionary containing rules for additional tags.
    """

    def __init__(self, author, project, **kwargs):
        """Initialize a Butler class.

        Arguments:
            author (str): The itch.io user that will submit a project.
            project (str): The project that the user will submit.

        Kwargs:
            exec (str): The path to the Butler executable file. Defaults to "butler".
        """
        self.author = author
        self.project = project
        self.tag_rules = DEFAULT_TAG_RULES.copy()
        self.bin = kwargs["exec"] if "exec" in kwargs else "butler"

    def push(self, file: str, **kwargs):
        """Push the file to the Itch.io page.

        Arguments:
            file (str): The path to the file to push.
            **kwargs: Arbitrary keyword arguments

        Kwargs:
            user_version (str): The user version to use, if any.
            with_tag_rule (str): The tag rule to use. This is used as a channel.
            with_custom_tag (str): The tag to use at the end of the tag rule name.
        """
        channel = self.author + "/" + self.project
        command = [self.bin, file, channel]

        if "with_tag_rule" in kwargs:
            channel += ":" + self.tag_rules.get(kwargs["with_tag_rule"])

        if "with_custom_tag" in kwargs:
            channel += "-" + kwargs["with_custom_tag"]

        if "user_version" in kwargs:
            command += ["--userversion", kwargs["user_version"]]
        return proc.check_call(command)

    def add_tag_rule(self, name: str, platform: ButlerPlatformType):
        """Create a new rule for the project's channel tags.

        Arguments:
            name (str): The rule that will determine what platform to store it under.
            platform (ButlerPlatformType): The platform for that rule.
        """
        self.tag_rules[name] = platform



"""
    This module contains a customized version of the `ZipFile` object
    to properly handle extraction of macOS apps so that permissions aren't
    stripped.
"""
import os
from zipfile import ZipFile as PyZipFile, ZipInfo
from renpy_distribute_tools.util import deprecated


class ZipFile(PyZipFile):
    """The ZipFile class is a patched version of the ZipFile class in the `zipfile` module that
        retains the attribute permissions for files.
    """

    def extract(self, member, path=None, pwd=None):
        if not isinstance(member, ZipInfo):
            member = self.getinfo(member)

        if path is None:
            path = os.getcwd()

        ret_val = self._extract_member(member, path, pwd)
        attr = member.external_attr >> 16
        if attr != 0:
            os.chmod(ret_val, attr)
        return ret_val

    def extractall(self, path=None, members=None, pwd=None):
        if members is None:
            members = self.namelist()

        if path is None:
            path = os.getcwd()
        else:
            path = os.fspath(path)

        for zipinfo in members:
            self.extract(zipinfo, path, pwd)


@deprecated("MyZipFile has been renamed to ZipFile.")
class MyZipFile(ZipFile):
    """The MyZipFile class is a patched version of the ZipFile class in the `zipfile` module that
        retains the attribute permissions for files.
    """



"""The Ren'Py Distribution Tools is a Python module that aims to make distribution of Ren'Py visual
    novel projects easier by automating it with Python utilities and classes.

## Getting Started

### Quick Start: Install via PyPI/Poetry

To install via PyPI:

```
pip install renpy-distribute-tools
```

Or, if you're using a Poetry project, just add the dependency:

```
poetry add renpy-distribute-tools
```

## What's included

The Ren'Py Distribution Tools set comes with utilities that make it easy to do the following:

- Modifying a visual novel's `Info.plist` with an identifier and copyright field.
- Code-signing the visual novel binaries in the Mac app with entitlements.
- Creating a ZIP copy of the Mac app and sending it to Apple's notarization servers.
- Verifying the status of a notarization request.
- Stapling the notarization ticket to a macOS app.
- Publishing content to Itch.io.
"""
__version__ = '0.4.0'

from renpy_distribute_tools.fixed_zipfile import ZipFile, MyZipFile
from renpy_distribute_tools.plist import fix_plist
from renpy_distribute_tools.apple import package_app_zip, \
    upload_to_notary, build_pkg, staple, code_sign, check_notary_status
from renpy_distribute_tools.util import deprecated, file_exists, find_mac_build, verify_built_files
from renpy_distribute_tools.itch import Butler, ButlerPlatformType, DEFAULT_TAG_RULES



"""The `apple` module contains important utilities to make signing, notarizing, and building
    packages for macOS easier.

The functions in this module require a device running macOS with Xcode 10 or higher.
"""
import os
import subprocess
from sys import platform
from functools import wraps


def darwin_only(call):
    """A decorator for macOS-specific commands.

    This should be used to denote that a function only works on macOS due to reliance on built-in
        tools from macOS or Xcode.
    """
    @wraps(call)
    def darwin_call(*args, **kwargs):
        if platform.lower() != "darwin":
            raise OSError("Function %s only works on macOS." % (call))
        return call(*args, **kwargs)
    return darwin_call


@darwin_only
def package_app_zip(app: str):
    """Create a ZIP file of the app.

    Args:
        app (str): The path to the macOS to make an archive of
    """
    if os.path.isdir(app):
        zip_commands = ["ditto", "-c", "-k", "--rsrc",
                        "--keepParent", app, app + ".zip"]
        subprocess.check_call(zip_commands)
    else:
        raise NotADirectoryError(
            "The .app file is either missing or not present.")


@darwin_only
def build_pkg(app: str, identity: str, package_name: str):
    """Create an installable package from a macOS app.

    By default, it will create an app package that installs to `/Applications/`. This package
        installer can also be used to submit an app to the Mac App Store.

    If the package name isn't a file path, `.pkg` will automatically be appended at the end of the
        name.

    Args:
        app (str): The path to the app to create a package of.
        identity (str): The identity to sign the package with
        package_name (str): The name or path of the resulting package.
    """
    package_file = package_name
    if ".pkg" not in package_name:
        package_file = package_name + ".pkg"
    commands = ["productbuild", "--component", app,
                "/Applications", "--sign", identity, package_file]
    return subprocess.check_call(commands)


@darwin_only
def code_sign(identity: str, app_directory: str, **kwargs):
    """Digitally sign a macOS application with a signing identity and any entitlements.

    Args:
        identity (str): The identity to use during signing, usually a Developer ID.
        app_directory (str): The path to the macOS application for signing.
        **kwargs: Arbitrary keyword arguments.

    Kwargs:
        entitlements (str): (Optional) The path to the entitlements the app should be signed with.
        enable_hardened_runtime (bool): Whether to sign the app with the hardened runtime on.
    """
    commands = ["codesign",
                "--timestamp",
                "--deep",
                "--force",
                "--no-strict",
                "--sign",
                identity,
                app_directory]

    if "entitlements" in kwargs:
        commands.append("--entitlements")
        commands.append(kwargs["entitlements"])

    if "enable_hardened_runtime" in kwargs and kwargs["enable_hardened_runtime"]:
        commands.append("--options=runtime")

    return subprocess.check_call(commands)


@darwin_only
def upload_to_notary(app: str,
                     identifier: str,
                     username: str,
                     password: str,
                     **kwargs) -> str:
    """Upload a macOS application archive to Apple's notary service for notarization.

    Args:
        app (str): The path to the macOS application to send to Apple.
        identifier (str): The bundle identifier of the application.
        username (str): The username (email address) of the Apple ID to notarize under.
        password (str): The password of the Apple ID to notarize under.
        **kwargs: Arbitrary keyword arguments.

    Kwargs:
        provider (str): The App Store Connect or iTunes Connect provider associated with the Apple
            ID used to sign the app.
    Returns:
        uuid_str (str): The request UUID.
    """
    package_app_zip(app)
    commands = ["xcrun", "altool", "-t", "osx", "-f", app + ".zip",
                "--notarize-app", "--primary-bundle-id", identifier,
                "-u", username, "-p", password]

    if "provider" in kwargs:
        commands += ["-itc_provider", kwargs["provider"]]

    result = subprocess.check_output(  # pylint:disable=unexpected-keyword-arg
        commands, text=True)
    os.remove(app + ".zip")

    result = result.split("\n")
    trimmed = result[1:]
    if len(trimmed) < 1:
        return ""
    return trimmed[0].replace("RequestUUID = ", "")


@darwin_only
def check_notary_status(uuid: str, username: str, password: str) -> int:
    """Get the notarization status of a given UUID.

    Arguments:
        uuid (str): The UUID of the app to check the status of.
        username (str): The user that submitted the notarization request.
        password (str): The password to use to sign into Apple.

    Returns:
        status (int): The status code associated with the UUID notarization request. A code of `-1`
            indicates that getting the status code failed, either because the item could not be
            found or because no status code has been given yet.
    """
    result = subprocess.check_output(  # pylint:disable=unexpected-keyword-arg
        ["xcrun", "altool", "--notarization-info", uuid, "-u", username, "-p", password], text=True)
    status = [x for x in result.replace(
        "  ", "").split("\n") if "Status Code" in x]

    if len(status) < 1:
        return -1

    status = status[0].replace("Status Code", "").replace(" ", "").split(":")
    while '' in status:
        status.remove('')

    if len(status) < 1:
        return -1
    return int(status[0])


@darwin_only
def staple(app: str):
    """Staple a notarization ticket to a notarized app.

    Args:
        app (str): The path of the macOS app to staple the ticket to.
    """
    commands = ["xcrun", "stapler", "staple", app]
    return subprocess.check_call(commands)



"""
    This module contains some utilities that help fix the property lists
    in the macOS versions of a Ren'Py project.
"""
import plistlib


def fix_plist(plist_file: str, identifier: str, p_copyright: str):
    """Add the bundle identifier and copyright text to a Ren'Py-built macOS app's Info.plist.

    Args:
        plist_file (str): The path to the plist file to modify
        identifier (str): The bundle identifier to set the app to
        p_copyright (str): The human-readable copyright text to attach
    """
    with open(plist_file, 'rb') as file_obj:
        properties = plistlib.load(file_obj)

    properties["CFBundleIdentifier"] = identifier
    properties["NSHumanReadableCopyright"] = p_copyright

    with open(plist_file, 'wb') as file:
        plistlib.dump(properties, file)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['renpy_distribute_tools']

package_data = \
{'': ['*']}

install_requires = \
['pylint>=2.4.4,<3.0.0']

setup_kwargs = {
    'name': 'renpy-distribute-tools',
    'version': '0.4.0',
    'description': "Utilities that make Ren'Py distribution less of a pain in the arse",
    'long_description': '# Ren\'Py Distribution Tools\n\nThe Ren\'Py Distribution Tools (RDT) is a set of utilities that make distribution of Ren\'Py projects easier and more seamless programmatically.\n\n[![MIT](https://img.shields.io/github/license/unscriptedvn/rdt)](LICENSE.txt)\n![Python](https://img.shields.io/badge/python-3.7+-blue.svg)\n[![PyPI version](https://badge.fury.io/py/renpy-distribute-tools.svg)](https://pypi.org/project/renpy-distribute-tools)\n\n## Getting Started\n\n### Quick Start: Install via PyPI/Poetry\n\nTo install via PyPI:\n\n```\npip install renpy-distribute-tools\n```\n\nOr, if you\'re using a Poetry project, just add the dependency:\n\n```\npoetry add renpy-distribute-tools\n```\n\n### Building from source\n\n#### Requirements\n\n- Python 3.7 or higher\n- Poetry package manager\n\nRen\'Py Distribute Tools is a Poetry project and can be built using Poetry\'s `build` command.\n\n1. Clone the repository.\n2. In the root of the project, run `poetry install`.\n3. Finally, run `poetry build`.\n\n## What\'s included\n\nThe Ren\'Py Distribution Tools set comes with utilities that make it easy to do the following:\n\n- Modifying a visual novel\'s `Info.plist`.\n- Code-signing the visual novel binaries in the Mac app with entitlements.\n- Creating a ZIP copy of the Mac app and sending it to Apple\'s notarization servers.\n- Verifying the notarization status of an app.\n- Stapling the notarization ticket to a macOS app.\n- Distributing app content through Itch.io.\n\n## Usage\n\nThe documentation website covers all of the modules contained in the RDT package. [View docs &rsaquo;](https://unscriptedvn.github.io/rdt/)\n\n### Example Usage: Notarizing a macOS build\n\n```py\nimport renpy_distribute_tools as rdt\n\nauthor = "Joe Smith"\nbundle_identifier = "com.example.my-vn"\ncode_sign_identity = "Developer ID Application: Joe Smith (XXXXXXXXXX)"\napp_path = "VN-1.0.0-dists/VN-1.0.0-mac/VN.app"\napple_email = "example.joesmith@icloud.com"\napple_provider = "XXXXXXXXXX"\n\nrdt.fix_plist(app_path + "/Contents/Info.plist",\n          bundle_identifier,\n          "\xc2\xa9 %s %s. All rights reserved." % (date.today().year, author))\nrdt.code_sign(code_sign_identity,\n          app_path,\n          entitlements="../../entitlements.plist",\n          enable_hardened_runtime=True)\ntry:\n    rdt.upload_to_notary(app_path,\n                     bundle_identifier,\n                     apple_email,\n                     "@keychain:AC_PASSWORD",\n                     provider=apple_provider)\nexcept CalledProcessError:\n    print("\\033[31;1mNotarization request failed. Aborting.\\033[0m")\n    sys.exit(1)\n```\n\n## License\n\nRDT is licensed under the MIT License, which is available in the LICENSE file in the source code repository.\n',
    'author': 'Marquis Kurt',
    'author_email': 'software@marquiskurt.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/UnscriptedVN/rdt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



"""This module contains some basic utilities to locate Ren'Py distributables and to find the macOS
    version of a VN.
"""
import os
import warnings
from typing import Optional, Callable
from functools import wraps


def deprecated(message: str = ""):
    """Mark a function as deprecated.

    This is used as a decorator to mark some functions as deprecated without needing to import
        warnings repeatedly. The function that uses the decorator will be called but will display
        a deprecation warning with a supplied message.

    Arguments:
        message (str): The message or reason for deprecation. Defaults to a generic statement
            generated by the function's name.

    Returns:
        warnable (Callable): The function with a warning wrapper.
    """
    def warnable(call: Callable):
        @wraps(call)
        def do_call(*args, **kwargs):
            warnings.warn(message if message else call.__name__ + " is deprecated.",
                          category=DeprecationWarning)
            call(*args, **kwargs)
        return do_call
    return warnable


@deprecated(message="Please use isfile or is from the os.path module.")
def file_exists(directory: str = os.getcwd(), item: str = '') -> Optional[str]:
    """Determine whether a file exists in a certain directory.

    **Note**: This function is being deprecated in favor of the utilities provided in the `os` 
        module.

    Args:
        dir (str): The directory to search in.
        item (str): The item to search for in the directory.
    Returns:
        fname (str): File name if it's found or `None` if it doesn't find anything
    """
    cwd = os.listdir(directory)
    for listed_item in cwd:
        if item in listed_item:
            return listed_item
    return None


@deprecated("This function is no longer supported.")
def verify_built_files(directory: str = os.getcwd()) -> Optional[str]:
    """Determine if the Ren'Py distributions have been built already by looking for the `-dists`
        directory.

    This function will check if the directory exists in itself.

    Args:
        dir (str): The directory to search.

    Returns:
        isdir (bool): Whether the directory exists or not.
    """
    return file_exists(directory=directory, item="-dists")


@deprecated("This function is longer supported.")
def find_mac_build(directory: str) -> bool:
    """Determine whether the macOS builds have been created.

    Args:
        dir (str): The directory to search in

    Returns:
        isfile (bool): Whether the macOS ZIP file exists.
    """
    return file_exists(directory=directory, item="-mac")
